require_relative 'ssh-publickey/version'

# Convert SSH public key from/to RFC4716/OpenSSH format
#
class SSHPublicKey
    # Report public key processing error
    class PublicKeyError < StandardError
    end

    
    # Maximum text size in a line for PEM format
    LINE_BREAK = 70

    
    # Regex for authorized key
    AUTHORIZED_KEY_REGEX =
        /^(?:    (?<directives>(?:[!-~]|\s(?=.*"))+) \s+ )?   # Options
                 (?<type>      [a-z0-9_-]+)          \s+      # Type
                 (?<key>       [A-Za-z0-9+\/]+=*)             # Key
          (?:\s+ (?<comment>.*?)                     \s* )?   # Comments
         $/x

    
    # Regex for OpenSSH public key
    # Same as authorized keys but without directives
    PUBKEY_OPENSSH_REGEX =
        /^       (?<type>   [a-z0-9_-]+)             \s+      # Type
                 (?<key>    [A-Za-z0-9+\/]+=*)                # Key
          (?:\s+ (?<comment>.*?)                     \s* )?   # Comments
         $/x

    
    # Regex for RFC4716 public key
    PUBKEY_RFC4716_REGEX =
        /^----\sBEGIN\sSSH2\sPUBLIC\sKEY\s----\R
          (?<tags> (?:[^:\p{Space}\p{Cntrl}]{1,64}            # Tags
                      \s*:\s*
                      (?:[^\\\r\n]*\\\R)* [^\\\r\n]+\R)*)
          (?<key>  (?:[A-Za-z0-9+\/]+\R)*                     # Key
                   [A-Za-z0-9+\/]+=*\R)
          ----\sEND\sSSH2\sPUBLIC\sKEY\s----
         $/xmu


    # :nodoc:
    PUBKEY_RFC4716_HEADERTAG_REGEX = 
        /[^:\p{Space}\p{Cntrl}]{1,64}
            \s*:\s*   
            (?:[^\\\r\n]*\\\R)*[^\\\r\n]+\R/xmu

    
    # Convert a public key in OpenSSH format to RFC4716 format
    #
    # @param  [String] pubkey   public key in OpenSSH format
    #
    # @return [String] public key in RFC4716 format
    #
    def self.openssh_to_rfc4716(pubkey)
        unless m = AUTHORIZED_KEY_REGEX.match(pubkey)
            raise PublicKeyError, "invalid OpenSSH public key"
        end

        # Lines are limited to 72 8-bytes char
        #  - limit ourselves to 70 to keep room for \\ and \n
        #  - comment part can be unicode so 1 char can be more that 1 byte
        linesize = 0
        comment  = "Comment: #{m[:comment]}"
                      .each_char.slice_before {|c|
                           bytesize = c.bytes.size 
                           if linesize + bytesize > LINE_BREAK
                           then linesize  = 0        ; true
                           else linesize += bytesize ; false
                           end
                      }.map(&:join).join("\\\n")
        key      = m[:key].scan(/.{1,70}/).join("\n")
        
        reskey   = []
        reskey   << "---- BEGIN SSH2 PUBLIC KEY ----"
        reskey   << comment
        reskey   << key
        reskey   << "---- END SSH2 PUBLIC KEY ----"
        reskey.join("\n")
    end

    
    # Convert a public key in RFC4716 format to OpenSSH format
    #
    # @param  [String] pubkey   public key in RFC4716 format
    #
    # @return [String] public key in OpenSSH format
    #
    def self.rfc4716_to_openssh(pubkey)
        unless m = PUBKEY_RFC4716_REGEX.match(pubkey) 
            raise PublicKeyError, "invalid RFC4716 public key"
        end

        key        = m[:key].gsub(/\R/, '')
        keydata    = m[:key].unpack1('m')
        len        = keydata.unpack1('N')
        type       = keydata[4,len]
        tags       = Hash[m[:tags].scan(PUBKEY_RFC4716_HEADERTAG_REGEX)
                                  .map {|tag| tag.gsub(/\\\R/, '').strip }
                                  .map {|tag| tag.split(/\s*:\s*/, 2)    }]
        comment    = tags.transform_keys {|k| k.downcase }['comment']
        directives = nil
        
        [directives, type,  key, comment ].compact.join(' ')
    end


    # Test if a public key is in OpenSSH format
    #
    # @param  [String] pubkey   public key in RFC4716 format
    #
    # @return [Boolean] 
    #
    def self.is_openssh?(pubkey)
        PUBKEY_RFC4716_REGEX.match?(pubkey)
    end

    
    # Test if a public key is in RFC4716 format
    #
    # @param  [String] pubkey   public key in RFC4716 format
    #
    # @return [Boolean]
    #
    def self.is_rfc4716?(pubkey)
        PUBKEY_RFC4716_REGEX.match?(pubkey)
    end
end
